# -*- coding: utf-8 -*-
"""WSGI middleware initialization for the algobowl application."""
from depot.manager import DepotManager

from algobowl.config.app_cfg import base_config
from algobowl.config.environment import load_environment

__all__ = ["make_app"]

# Use base_config to setup the necessary PasteDeploy application factory.
# make_base_app will wrap the TG2 app with all the middleware it needs.
make_base_app = base_config.setup_tg_wsgi_app(load_environment)


def make_app(global_conf, full_stack=True, **app_conf):
    """
    Set algobowl up with the settings found in the PasteDeploy configuration
    file used.

    :param global_conf: The global settings for algobowl (those
        defined under the ``[DEFAULT]`` section).
    :type global_conf: dict
    :param full_stack: Should the whole TG2 stack be set up?
    :type full_stack: str or bool
    :return: The algobowl application with all the relevant middleware
        loaded.

    This is the PasteDeploy factory for the algobowl application.

    ``app_conf`` contains all the application-specific settings (those defined
    under ``[app:main]``.
    """
    app = make_base_app(global_conf, full_stack=True, **app_conf)
    app = DepotManager.make_middleware(app)

    return app



import tg
from repoze.who.api import get_api
from tg import expose, flash, lurl, predicates, redirect, require, tmpl_context, url
from tg.exceptions import HTTPFound
from tgext.admin.controller import AdminController

import algobowl.controllers.file_redirector as file_redirector
import algobowl.controllers.pref as pref_controller
import algobowl.controllers.setup as setup_controller
import algobowl.model
from algobowl.config.app_cfg import AdminConfig
from algobowl.controllers import api
from algobowl.controllers.competition import CompetitionsController
from algobowl.controllers.error import ErrorController
from algobowl.controllers.group import GroupsController
from algobowl.lib.base import BaseController
from algobowl.model import DBSession

__all__ = ["RootController"]


class RootController(BaseController):
    admin = AdminController(algobowl.model, DBSession, config_type=AdminConfig)
    api = api.ApiController()
    setup = setup_controller.SetupController()
    pref = pref_controller.PrefController()
    error = ErrorController()
    group = GroupsController()
    competition = CompetitionsController()

    def _before(self, *args, **kw):
        tmpl_context.project_name = "algobowl"

    @expose("algobowl.templates.index")
    def index(self):
        """Handle the front-page."""
        return dict(page="index")

    @expose()
    def algobowl(self):
        """Redirect for old route to homepage"""
        redirect(url("/"))

    @expose("algobowl.templates.privacy")
    def privacy(self):
        return dict(page="privacy")

    @expose("algobowl.templates.tos")
    def tos(self):
        return dict(page="tos")

    @expose()
    def files(self, filename):
        file = file_redirector.get_file(filename)
        tg.response.content_type = "application/octet-stream"
        return file.file.read()

    @expose()
    def login(self):
        if not tg.request.identity:
            tg.request.environ["repoze.who.challenge"] = "mpapi"
            who_api = get_api(tg.request.environ)
            return who_api.challenge()
        redirect(url("/"))

    @expose()
    def glogin(self):
        if not tg.request.identity:
            tg.request.environ["repoze.who.challenge"] = "glogin"
            who_api = get_api(tg.request.environ)
            return who_api.challenge()
        redirect(url("/"))

    @expose()
    @require(predicates.not_anonymous())
    def logout(self):
        who_api = get_api(tg.request.environ)
        headers = who_api.logout()
        return HTTPFound(headers=headers)

    @expose()
    def post_login(self, came_from=lurl("/")):
        if tg.request.identity:
            user = tg.request.identity["user"]
            flash("Welcome, {}!".format(user), "success")
            u = tg.request.relative_url(str(came_from), to_application=True)
            if not u.startswith(tg.request.application_url):
                flash("Dangerous redirect prevented", "warning")
                redirect("/")
            redirect(u)
        else:
            flash("Login failure", "error")
            redirect("/")

    @expose()
    @require(predicates.has_permission("admin"))
    def force_error(self):
        raise RuntimeError("Forced error")



# -*- coding: utf-8 -*-
"""Setup the algobowl application"""
from __future__ import print_function

import transaction
from tg import config


def setup_schema(command, conf, vars):
    """Place any commands to setup algobowl here"""
    # Load the models

    # <websetup.websetup.schema.before.model.import>
    from algobowl import model

    # <websetup.websetup.schema.after.model.import>
    # <websetup.websetup.schema.before.metadata.create_all>
    print("Creating tables")
    model.metadata.create_all(bind=config["tg.app_globals"].sa_engine)
    # <websetup.websetup.schema.after.metadata.create_all>
    transaction.commit()
    print("Initializing Migrations")
    import alembic.config

    alembic_cfg = alembic.config.Config()
    alembic_cfg.set_main_option("script_location", "migration")
    alembic_cfg.set_main_option("sqlalchemy.url", config["sqlalchemy.url"])
    import alembic.command

    alembic.command.stamp(alembic_cfg, "head")



import datetime
import re

import tg

import algobowl.model as model


def input_redirector(group_id):
    group_id = int(group_id)
    group = model.DBSession.query(model.Group).get(group_id)
    if not group:
        tg.abort(404, "No such group.")
    competition = group.competition

    user = tg.request.identity and tg.request.identity.get("user")
    is_admin = user and user.admin
    is_group_member = user in group.users
    file_is_public = datetime.datetime.now() >= competition.output_upload_begins

    if not file_is_public and not is_admin and group.input and group.input.is_default:
        tg.abort(403, "Default inputs are not downloadable until output upload begins.")

    if file_is_public or is_group_member or is_admin:
        if not group.input:
            tg.abort(404, "An input file has not been uploaded yet.")
        return group.input.data

    tg.abort(403, "You do not have access to this file at this time.")


def output_redirector(from_group_id, to_group_id):
    from_group_id = int(from_group_id)
    to_group_id = int(to_group_id)
    output = (
        model.DBSession.query(model.Output)
        .join(model.Output.input)
        .filter(model.Input.group_id == to_group_id)
        .filter(model.Output.group_id == from_group_id)
        .filter(model.Output.active == True)
        .one()
    )
    if not output:
        tg.abort(404, "Output does not exist.")

    competition = output.group.competition
    user = tg.request.identity and tg.request.identity.get("user")
    is_admin = user and user.admin
    is_uploader = user in output.group.users
    is_verifier = user in output.input.group.users
    visible_to_verifier = datetime.datetime.now() >= competition.verification_begins
    file_is_public = competition.open_verification_open

    if (
        file_is_public
        or is_uploader
        or is_admin
        or (is_verifier and visible_to_verifier)
    ):
        return output.data

    tg.abort(403, "You do not have access to this file at this time.")


file_pattern_handlers = [
    (re.compile(r"input_group(\d+)"), input_redirector),
    (re.compile(r"output_from_(\d+)_to_(\d+)"), output_redirector),
]


def get_file(filename):
    for pattern, func in file_pattern_handlers:
        m = pattern.fullmatch(filename)
        if m:
            return func(*m.groups())

    tg.abort(404, "No handler found for filename.")



"""
Problem Format Helpers
======================

This module contains two things:

1. The BaseInput and BaseOutput classes, which define the base classes
   for input classes and output classes respectively.

2. Any helper functions to help implement problems.
"""

import dataclasses
import enum
import pathlib
import random
import types
import typing

from typing_extensions import Self

from algobowl.lib import problem_tester


class FileFormatError(Exception):
    """Raised from .read() for either an input or an output.

    This exception will always notify the user of the formatting error
    immediately.  The exception message will be displayed to the user.
    """


class VerificationError(Exception):
    """Raised from .verify() (or .compute_actual_score()) of an output.

    This exception will not notify the user, and will count as an
    output rejection from the instructor.  The instructor will be able
    to see the exception message, but not students.
    """


class ProblemTestError(Exception):
    """Tests did not pass successfully."""


class RankSort(enum.Enum):
    minimization = 0
    maximization = 1


@dataclasses.dataclass
class BaseInput:
    @classmethod
    def read(cls, f):
        """Read an input file."""
        raise NotImplementedError

    def write(self, f):
        """Write an input file."""
        raise NotImplementedError

    @classmethod
    def generate(cls, rng: random.Random) -> Self:
        """Generate an input."""
        raise NotImplementedError

    def trivial_solve(self) -> "BaseOutput":
        """Solve this input in the most trivial way possible.

        This should solve the input in a valid but computationally cheap manner.
        Perhaps you want to aim to solve the problem with the worst answer
        possible.  Problems are not required to implement this, but it does
        provide some fuzz testing of outputs and the verifier when combined with
        the generate() method.
        """
        raise NotImplementedError


@dataclasses.dataclass
class BaseOutput:
    input: BaseInput
    score: int
    rank_sort: typing.ClassVar[RankSort] = RankSort.minimization

    @classmethod
    def read(cls, input, f):
        """Read an output file."""
        raise NotImplementedError

    def compute_actual_score(self):
        """Compute the correct score that should correspond to this output.

        Returns:
            The integer score value.

        Raises:
            A VerificationError, if the output is not valid.
        """
        raise NotImplementedError

    def verify(self):
        """Verify an output is valid.

        Raises:
            A VerificationError, if the output is not valid.
        """
        actual_score = self.compute_actual_score()
        if self.score != actual_score:
            raise VerificationError(
                f"Reported score ({self.score}) does not match actual ({actual_score})"
            )

    @staticmethod
    def repr_score(score):
        """Get the score as a string."""
        return str(score)

    def write(self, f):
        """Write an output."""
        raise NotImplementedError


def check_bound(lineno, bound, value):
    if bound is None:
        return
    if isinstance(bound, int):
        cond = value == bound
    elif callable(bound):
        cond = bound(value)
    else:
        cond = value in bound
    if cond:
        return
    raise FileFormatError(f"Line #{lineno + 1}: Value {value} out of expected bounds")


def assert_linecount(lines, bounds):
    try:
        check_bound(-1, bounds, len(lines))
    except FileFormatError:
        # Try to be accommodating of extra blank lines at end of file.
        while lines and not lines[-1].rstrip():
            lines.pop()
            try:
                check_bound(-1, bounds, len(lines))
                return
            except FileFormatError:
                continue
        raise FileFormatError("File has invalid number of lines")


def parse_int(lineno, value, bound=None):
    try:
        value = int(value)
    except ValueError:
        raise FileFormatError(
            f"Line #{lineno + 1}: Unable to interpret {value!r} as integer"
        )
    check_bound(lineno, bound, value)
    return value


def parse_line_ints(lines, lineno, bounds=None, count=None):
    line = lines[lineno]
    line_split = line.split()
    try:
        check_bound(lineno, count, len(line_split))
    except FileFormatError:
        raise FileFormatError(
            f"Line #{lineno + 1}: Number of values ({len(line_split)}) "
            f"is out of expected bounds"
        )
    return [parse_int(lineno, value, bounds) for value in line_split]


def parse_line_int(lines, lineno, bounds=None):
    return parse_line_ints(lines, lineno, bounds=bounds, count=1)[0]


_module_cache = {}
_statement_cache = {}


def _cache_getent(cache, path, loader=lambda pth: pth.read_bytes()):
    path = path.resolve()
    mtime = path.stat().st_mtime
    if path in cache:
        c_mtime, c_val = cache[path]
        if c_mtime == mtime:
            return c_val
    val = loader(path)
    cache[path] = (mtime, val)
    return val


class Problem:
    """High-level wrapper around the problem format."""

    def __init__(self, path):
        self.path = path

    def get_module(self):
        def loader(path):
            contents = path.read_text()
            module = types.ModuleType("problem")
            code = compile(contents, str(path), "exec")
            exec(code, module.__dict__)
            return module

        return _cache_getent(_module_cache, self.path / "problem.py", loader=loader)

    def get_statement_pdf(self):
        return _cache_getent(_statement_cache, self.path / "statement.pdf")

    def generate_input(self, rng):
        module = self.get_module()
        return module.Input.generate(rng)

    def parse_input(self, input_file):
        module = self.get_module()
        return module.Input.read(input_file)

    def parse_output(self, input, output_file):
        module = self.get_module()
        if not isinstance(input, module.Input):
            input = module.Input.read(input)
        output = module.Output.read(input, output_file)
        return output

    def verify_output(self, input, output_file):
        output = self.parse_output(input, output_file)
        output.verify()

    def run_tests(self, pytest_extra_args=()):
        """Run tests on this problem."""
        retcode = problem_tester.run_problem_tests(self.path, pytest_extra_args)
        if retcode != 0:
            raise ProblemTestError(f"Tests failed with error code {retcode}")


class DefaultProblem(Problem):
    """Used when the problem is unspecified."""

    def __init__(self):
        pass

    def get_module(self):
        module = types.ModuleType("problem")
        module.Input = BaseInput
        module.Output = BaseOutput
        return module


def load_problem(competition):
    # Deferred import so this module does not depend on TurboGears for
    # most functionality.
    import tg

    if not competition.problem:
        return DefaultProblem()

    for path in tg.config.get("problems.search_paths", "").split():
        if not path:
            continue
        path = pathlib.Path(path)
        if (path / competition.problem).is_dir():
            return Problem(path / competition.problem)
    raise ValueError(f"No problem named {competition.problem}")



import datetime
import zipfile
from collections import defaultdict, namedtuple
from io import BytesIO, StringIO

from recordclass import recordclass
from sqlalchemy.sql.expression import case
from tg import abort, expose, flash, redirect, request, require, response
from tg.predicates import has_permission

import algobowl.lib.problem as problemlib
from algobowl.lib.base import BaseController
from algobowl.lib.logoutput import logoutput
from algobowl.model import (
    Competition,
    DBSession,
    Evaluation,
    Group,
    Input,
    Output,
    Protest,
    VerificationStatus,
)

__all__ = ["CompetitionsController", "CompetitionController"]


ScoreTuple = namedtuple(
    "ScoreTuple", ["score", "verification", "rank", "output", "vdiffer"]
)

GradingTuple = recordclass(
    "GradingTuple",
    ["rankings", "fleet", "verification", "input_ones", "contributions", "evaluations"],
)

GradingContributionTuple = recordclass(
    "GradingContributionTuple",
    ["participation", "verification", "input_submitted", "input_difficulty", "ranking"],
    defaults=[0, 0, 0, 0, 0],
)

GradingVerificationTuple = recordclass(
    "GradingVerificationTuple",
    ["correct", "false_positives", "false_negatives"],
    defaults=[0, 0, 0],
)

CompetitionYearTuple = recordclass("CompetitionYearTuple", ["year", "competitions"])


class GroupEntry:
    def __init__(self):
        self.reject_count = 0
        self.sum_of_ranks = 0
        self.penalties = 0
        self.place = 1
        self.input_ranks = {}

    @property
    def score(self):
        return self.sum_of_ranks + self.penalties

    @property
    def adj_score(self):
        """
        Score used for grading: each rejection is assumed
        to count as the number of groups.
        """
        if not self.reject_count:
            return self.score

        num_inputs = len(self.input_ranks)
        if not num_inputs:
            num_inputs = 9999

        return self.sum_of_ranks + self.penalties + self.reject_count * num_inputs

    def __lt__(self, other):
        if self.reject_count < other.reject_count:
            return True
        if self.reject_count > other.reject_count:
            return False
        return self.score < other.score

    def __eq__(self, other):
        return self.reject_count == other.reject_count and self.score == other.score

    def to_dict(self):
        return {
            "reject_count": self.reject_count,
            "sum_of_ranks": self.sum_of_ranks,
            "penalties": self.penalties,
            "score": self.score,
            "place": self.place,
            "input_ranks": {
                k.id: (v[0], str(v[1]), v[2]) for k, v in self.input_ranks.items()
            },
        }


def compute_rankings_grade(gt, fleet_num, fleet):
    place_in_fleet = 0
    last_adj_score = fleet[0].rankings.adj_score
    for i, other_gt in enumerate(fleet):
        if other_gt.rankings.adj_score != last_adj_score:
            place_in_fleet = i
            last_adj_score = other_gt.rankings.adj_score
        if other_gt is gt:
            break
    if len(fleet) == 1:
        return 5 + fleet_num * 5
    return (place_in_fleet / (len(fleet) - 1)) * 5 + fleet_num * 5


class CompetitionController(BaseController):
    def __init__(self, competition):
        self.competition = competition

    @expose("algobowl.templates.competition.rankings")
    @expose("json")
    def index(self, ground_truth=False, incognito=False):
        user = request.identity and request.identity["user"]
        now = datetime.datetime.now()
        admin = user and user.admin
        comp = self.competition
        problem = problemlib.load_problem(comp)
        show_scores = admin or (
            comp.open_verification_begins and now >= comp.open_verification_begins
        )
        if user:
            my_groups = (
                DBSession.query(Group).filter(Group.competition_id == comp.id).all()
            )
            # non-sql filter, bleh
            my_groups = [g for g in my_groups if user in g.users]
        else:
            my_groups = []

        show_input_downloads = admin or not comp.archived
        show_incognito_option = admin or any(g.incognito for g in my_groups)

        if show_incognito_option:
            incognito_teams = (
                DBSession.query(Group)
                .filter(Group.competition_id == comp.id)
                .filter(Group.incognito == True)
                .all()
            )
        else:
            incognito = False
            incognito_teams = []

        if not admin and now < comp.output_upload_begins:
            flash("Rankings are not available yet", "info")
            redirect("/competition")

        if ground_truth and not admin:
            abort(403, "You do not have permission for this option")

        if ground_truth:
            verification_column = Output.ground_truth
        else:
            verification_column = case(
                [(Output.use_ground_truth, Output.ground_truth)],
                else_=Output.verification,
            )

        open_verification = user and comp.open_verification_open

        problem_module = problem.get_module()
        score_sort = {
            problemlib.RankSort.minimization: Output.score.asc(),
            problemlib.RankSort.maximization: Output.score.desc(),
        }[problem_module.Output.rank_sort]

        groups = defaultdict(GroupEntry)
        ir_query = (
            DBSession.query(Input, Group, Output, verification_column)
            .join(Input.outputs)
            .join(Output.group)
            .filter(Group.competition_id == comp.id)
            .filter(Output.active == True)
            .order_by(
                Input.group_id,
                verification_column == VerificationStatus.rejected,
                score_sort,
            )
        )

        inputs = []
        last_iput = None
        accurate_count = 0
        total_count = 0
        for iput, ogroup, output, verif in ir_query:
            if not incognito and ogroup.incognito:
                continue
            if iput is not last_iput:
                inputs.append(iput)
                potential_rank = 1
                last_rank = 0
                last_score = None
            shown_score = None
            shown_output = None
            if show_scores:
                shown_score = problem_module.Output.repr_score(output.score)
                shown_output = output
            if verif is VerificationStatus.rejected:
                rank = None
                shown_score = None
            elif output.score == last_score:
                rank = last_rank
            else:
                rank = potential_rank

            vdiffer = False
            if ground_truth:
                if (
                    output.use_ground_truth
                    or output.verification == output.ground_truth
                ):
                    accurate_count += 1
                else:
                    vdiffer = True

            groups[ogroup].input_ranks[iput] = ScoreTuple(
                shown_score, verif, rank, shown_output, vdiffer
            )
            if verif is VerificationStatus.rejected:
                groups[ogroup].reject_count += 1
            else:
                groups[ogroup].sum_of_ranks += rank

            # add accepted resubmissions to penalty
            if verif is VerificationStatus.accepted and not output.original:
                groups[ogroup].penalties += 1

            total_count += 1
            potential_rank += 1
            last_iput = iput
            last_rank = rank
            last_score = output.score

        # no submission? this adds to reject count
        for group in groups.values():
            for iput in inputs:
                if iput not in group.input_ranks.keys():
                    group.reject_count += 1

        # add open verification protest rejections to penalty
        rprotests = (
            DBSession.query(Protest)
            .filter(Protest.accepted == False)
            .join(Protest.submitter)
            .filter(Group.competition_id == comp.id)
        )
        for protest in rprotests:
            # technically, a group which failed to submit anything
            # COULD protest... but this case is unlikely ;)
            if protest.submitter in groups.keys():
                groups[protest.submitter].penalties += 1

        # compute places for groups
        for this_ent in groups.values():
            for other_ent in groups.values():
                if other_ent < this_ent:
                    this_ent.place += 1

        if request.response_type == "application/json":
            return {
                "status": "success",
                "groups": {k.id: v.to_dict() for k, v in groups.items()},
            }
        else:
            return {
                "groups": groups,
                "my_groups": my_groups,
                "competition": comp,
                "inputs": inputs,
                "admin": admin,
                "incognito": incognito,
                "show_incognito_option": show_incognito_option,
                "incognito_teams": incognito_teams,
                "show_input_downloads": show_input_downloads,
                "ground_truth": ground_truth,
                "verification_accuracy": (
                    0 if not total_count else (accurate_count / total_count)
                ),
                "open_verification": open_verification,
            }

    @expose("algobowl.templates.competition.grade")
    @require(has_permission("admin"))
    def grade(self):
        rankings = self.index(ground_truth=True, incognito=False)

        def new_gt(rankings_entry):
            return GradingTuple(
                rankings_entry,
                None,
                GradingVerificationTuple(),
                0,
                GradingContributionTuple(),
                defaultdict(dict),
            )

        groups = {k: new_gt(v) for k, v in rankings["groups"].items()}

        num_inputs = len(rankings["inputs"])
        benchmark_groups = []

        # in the case a group submitted an input but has no outputs
        # uploaded yet, they won't be in groups as they are off the
        # rankings table. in this case, we need to make a
        # GradingTuples for them now.
        for group in self.competition.groups:
            if group.incognito:
                groups.pop(group, None)
                continue
            if group.benchmark:
                benchmark_groups.append(groups[group])
                groups.pop(group, None)
                continue
            if group not in groups.keys():
                rankings_entry = GroupEntry()
                # If they submitted nothing, then everything is a "reject"
                rankings_entry.reject_count = num_inputs
                groups[group] = new_gt(rankings_entry)

        benchmark_groups.sort(key=lambda gt: -gt.rankings.adj_score)
        fleets = [[] for _ in range(len(benchmark_groups) + 1)]

        for group, gt in groups.items():
            # compute verification correctness
            q = (
                DBSession.query(Output)
                .filter(Output.original == True)
                .join(Output.input)
                .filter(Input.group_id == group.id)
            )
            for oput in q:
                if oput.verification == oput.ground_truth:
                    gt.verification.correct += 1
                elif oput.verification == VerificationStatus.accepted:
                    gt.verification.false_positives += 1
                else:
                    gt.verification.false_negatives += 1

            # Compute input difficulty
            for iput, st in gt.rankings.input_ranks.items():
                if iput.group.incognito or iput.group.benchmark:
                    continue
                if st.rank == 1:
                    groups[iput.group].input_ones += 1

            gt.fleet = 0
            for bench_gt in benchmark_groups:
                if gt.rankings.adj_score <= bench_gt.rankings.adj_score:
                    gt.fleet += 1
            fleets[gt.fleet].append(gt)

        for fleet in fleets:
            fleet.sort(key=lambda gt: -gt.rankings.adj_score)

        for group, gt in groups.items():
            gt.contributions.ranking = compute_rankings_grade(
                gt, gt.fleet, fleets[gt.fleet]
            )
            gt.contributions.verification = (
                (gt.verification.correct / sum(gt.verification) * 20)
                if any(gt.verification)
                else 0
            )
            gt.contributions.participation = (
                (num_inputs - gt.rankings.reject_count) / num_inputs * 50
            )
            if group.input and not group.input.is_default:
                gt.contributions.input_submitted = 5
                gt.contributions.input_difficulty = max(
                    10 - (gt.input_ones - 1),
                    0,
                )
            else:
                gt.contributions.input_submitted = 0
                gt.contributions.input_difficulty = 0

        for group, gt in groups.items():
            for from_member in group.users:
                q = (
                    DBSession.query(Evaluation, Evaluation.score)
                    .filter(Evaluation.group_id == group.id)
                    .filter(Evaluation.from_student_id == from_member.id)
                    .all()
                )
                evals = {e.to_student: s for e, s in q}
                for to_member in group.users:
                    if to_member not in evals.keys():
                        evals[to_member] = 1.0
                for to_member, score in evals.items():
                    gt.evaluations[to_member][from_member] = score / sum(evals.values())

        return {"groups": groups, "competition": self.competition}

    @logoutput
    @require(has_permission("admin"))
    def reverify(self):
        problem = problemlib.load_problem(self.competition)

        changes = 0
        for group in self.competition.groups:
            if group.input:
                for output in group.input.outputs:
                    old_status = output.ground_truth
                    input_file = StringIO(group.input.data.file.read().decode("utf-8"))
                    output_file = StringIO(output.data.file.read().decode("utf-8"))
                    try:
                        problem.verify_output(input_file, output_file)
                    except problemlib.VerificationError as e:
                        output.ground_truth = VerificationStatus.rejected
                        print("{} rejected because: {}".format(output, e))
                    except Exception as e:
                        output.ground_truth = VerificationStatus.waiting
                        print("Verifier module failed on {}: {}".format(output, e))
                    else:
                        output.ground_truth = VerificationStatus.accepted
                    if old_status != output.ground_truth:
                        print(
                            "{} changed ground truth from {} to {}".format(
                                output, old_status, output.ground_truth
                            )
                        )
                        changes += 1
        DBSession.flush()
        print("{} ground truths changed".format(changes))

    @expose("algobowl.templates.competition.ov")
    def ov(self, output_id):
        output = DBSession.query(Output).get(int(output_id))
        if output.group.competition_id != self.competition.id:
            abort(404)
        if not output.active:
            abort(404, "This output is no longer active.")
        user = request.identity and request.identity["user"]
        if user:
            group = (
                DBSession.query(Group)
                .filter(Group.competition_id == self.competition.id)
                .filter(Group.users.any(id=user.id))
                .first()
            )
        else:
            group = None

        problem = problemlib.load_problem(self.competition)
        problem_module = problem.get_module()
        show_input_downloads = (
            user and user.admin
        ) or not output.group.competition.archived
        message = request.POST.get("message")
        if group and message:
            if output.use_ground_truth:
                abort(403, "The instructor has already reviewed this output.")

            output.use_ground_truth = True
            protest = Protest(
                message=message,
                accepted=output.verification != output.ground_truth,
                submitter=group,
                output=output,
            )

            DBSession.add(output)
            DBSession.add(protest)

            flash("Your protest has been submitted.", "success")

        return {
            "output": output,
            "score": problem_module.Output.repr_score(output.score),
            "group": group,
            "show_input_downloads": show_input_downloads,
            "competition": self.competition,
        }

    @expose()
    def all_inputs(self):
        user = request.identity and request.identity["user"]
        now = datetime.datetime.now()
        comp = self.competition
        if not (user and user.admin) and now < comp.output_upload_begins:
            abort(
                403,
                "Input downloading is not available until the output"
                " upload stage begins.",
            )
        if not (user and user.admin) and comp.archived:
            abort(
                403,
                "Input downloading is unavailable for old competitions. "
                "Contact the site administrator if you need access.",
            )
        f = BytesIO()
        archive = zipfile.ZipFile(f, mode="w")
        inputs = (
            DBSession.query(Input)
            .join(Input.group)
            .filter(Group.competition_id == comp.id)
        )
        for iput in inputs:
            archive.writestr(
                "inputs/{}".format(iput.data.filename), iput.data.file.read()
            )
        archive.close()
        f.seek(0)
        response.content_type = "application/zip"
        return f.read()

    @expose()
    def problem_statement(self):
        user = request.identity and request.identity["user"]
        if not (user and user.admin) and self.competition.archived:
            abort(
                403,
                "The problem statement is no longer available as the "
                "competition has been archived.",
            )
        problem = problemlib.load_problem(self.competition)
        response.content_type = "application/pdf"
        return problem.get_statement_pdf()


class CompetitionsController(BaseController):
    @expose()
    def _lookup(self, comp_id, *args):
        try:
            comp_id = int(comp_id)
        except ValueError:
            abort(404, "Invalid input for competition id")
        competition = DBSession.query(Competition).get(comp_id)
        if not competition:
            abort(404, "No such competition")

        return CompetitionController(competition), args

    @expose("algobowl.templates.competition.archive")
    def archive(self):
        now = datetime.datetime.now()
        comps = []
        for comp in (
            DBSession.query(Competition)
            .filter(Competition.open_verification_ends <= now)
            .order_by(Competition.input_upload_ends.desc())
        ):
            if not comps or comps[-1].year != comp.input_upload_ends.year:
                comps.append(CompetitionYearTuple(comp.input_upload_ends.year, [comp]))
            else:
                comps[-1].competitions.append(comp)

        return {"competitions": comps}

    @expose("algobowl.templates.competition.list")
    def index(self):
        now = datetime.datetime.now()
        comps = (
            DBSession.query(Competition)
            .filter(Competition.input_upload_begins <= now)
            .filter(Competition.open_verification_ends > now)
            .order_by(Competition.input_upload_ends)
            .all()
        )
        if len(comps) == 1:
            redirect("/competition/{}".format(comps[0].id))
        return {"competitions": comps}



"""Setup the algobowl application"""

from algobowl.config.environment import load_environment

from .bootstrap import bootstrap
from .schema import setup_schema


def setup_app(command, conf, vars):
    load_environment(conf.global_conf, conf.local_conf)
    setup_schema(command, conf, vars)
    bootstrap(command, conf, vars)


__all__ = ["setup_app"]



from tg import expose, request

from algobowl.lib.base import BaseController

__all__ = ["ErrorController"]


class ErrorController(BaseController):
    """
    Generates error documents as and when they are required.

    The ErrorDocuments middleware forwards to ErrorController when error
    related status codes are returned from the application.

    This behaviour can be altered by changing the parameters to the
    ErrorDocuments middleware in your config/middleware.py file.

    """

    @expose("algobowl.templates.error")
    @expose("json")
    def document(self, *args, **kwargs):
        """Render the error document"""
        resp = request.environ.get("tg.original_response")
        try:
            # tg.abort exposes the message as .detail in response
            message = resp.detail
        except Exception:
            message = None

        if hasattr(resp, "status_int"):
            code = resp.status_int
        else:
            code = 404

        if not message:
            message = "We're sorry but we weren't able to process this request."

        values = dict(
            prefix=request.environ.get("SCRIPT_NAME", ""),
            code=request.params.get("code", code),
            message=request.params.get("message", message),
        )
        return values



import os
from urllib.parse import urlencode

import google_auth_oauthlib.flow as gflow
import googleapiclient.discovery as gdiscovery
import requests
import tg
import transaction
from repoze.who.interfaces import IAuthenticator, IChallenger, IIdentifier
from tg.configuration.auth import TGAuthMetadata
from tg.exceptions import HTTPFound
from webob import Request
from zope.interface import implementer

from algobowl.model import AuthToken, DBSession, User


class BaseAuth:
    def _get_cookie_plugin(self, environ):
        return environ["repoze.who.plugins"]["cookie"]

    def remember(self, environ, identity):
        return self._get_cookie_plugin(environ).remember(environ, identity)

    def forget(self, environ, identity):
        return self._get_cookie_plugin(environ).forget(environ, identity)


@implementer(IIdentifier, IAuthenticator)
class TokenAuth:
    def identify(self, environ):
        auth_header = environ.get("HTTP_AUTHORIZATION")
        if not auth_header:
            return None
        method, _, token = auth_header.partition(" ")
        if method != "Bearer":
            return None
        environ["tg.skip_auth_challenge"] = True
        return {"token": token, "identifier": "token"}

    def remember(self, environ, identity):
        return []

    def forget(self, environ, identity):
        return []

    def authenticate(self, environ, identity):
        if identity.get("identifier") != "token":
            return None
        client_id = identity["token"]

        token = (
            DBSession.query(AuthToken)
            .filter(AuthToken.client_id == client_id)
            .one_or_none()
        )
        if not token:
            return None

        return token.user.username


@implementer(IIdentifier, IChallenger, IAuthenticator)
class MPAPIAuthenticator(BaseAuth):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.s = requests.Session()

    def identify(self, environ):
        request = Request(environ)
        try:
            ticket = request.GET["tkt"]
        except KeyError:
            return None

        return {"login": ticket, "identifier": "mpapi"}

    def forget(self, environ, identity):
        headers = super().forget(environ, identity)
        headers.append(("Location", self.mpapi_slo))
        return headers

    def authenticate(self, environ, identity):
        try:
            ticket = identity["login"]
        except KeyError:
            return None

        if identity["identifier"] != "mpapi":
            return None

        r = requests.post(self.mpapi_fetch, data={"tkt": ticket})
        r.raise_for_status()
        data = r.json()
        if data["result"] != "success":
            raise ValueError("MPAPI Failure")
        username = data["uid"]
        attributes = data["attributes"]
        uid = attributes["uidNumber"]
        full_name = "{} {}".format(attributes["first"], attributes["sn"])
        email = attributes["mail"]

        user = DBSession.query(User).filter(User.id == uid).one_or_none()
        if user:
            # Update attributes in case of username/full name change
            user.username = username
            user.full_name = full_name
            user.email = email
        else:
            user = User(id=uid, username=username, full_name=full_name, email=email)
            DBSession.add(user)

        DBSession.flush()
        transaction.commit()
        return username

    def new_user_by_username(self, username):
        url = f"{self.mpapi_url}/uid/{username}"
        r = requests.get(url)
        r.raise_for_status()
        data = r.json()
        if data["result"] != "success":
            raise ValueError("MPAPI Failure.  User {username} may not exist?")
        attributes = data["attributes"]
        uid = attributes["uidNumber"]
        full_name = f"{attributes['first']} {attributes['sn']}"
        email = attributes["mail"]
        user = User(id=uid, username=username, full_name=full_name, email=email)
        DBSession.add(user)
        return user

    @property
    def mpapi_url(self) -> str:
        return tg.config["auth.mpapi.url"].rstrip("/")

    @property
    def mpapi_sso(self) -> str:
        return self.mpapi_url + "/sso"

    @property
    def mpapi_fetch(self) -> str:
        return self.mpapi_url + "/fetch"

    @property
    def mpapi_slo(self) -> str:
        return self.mpapi_url + "/slo"

    def challenge(self, environ, status, app_headers, forget_headers):
        """
        Provide ``IChallenger`` interface.
        """
        challenger = environ.get("repoze.who.challenge")
        if challenger and challenger != "mpapi":
            return None
        request = Request(environ)
        return_url = tg.url(
            request.application_url + "/post_login", {"came_from": request.path_qs}
        )
        headers = [
            (
                "Location",
                "{}?{}".format(self.mpapi_sso, urlencode({"return": return_url})),
            ),
            *forget_headers,
            *((h, v) for h, v in app_headers if h.lower() == "set-cookie"),
        ]
        return HTTPFound(headers=headers)


@implementer(IIdentifier, IChallenger, IAuthenticator)
class GoogleAuth(BaseAuth):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._flow = None

    @property
    def flow(self):
        if not self._flow:
            client_secrets_file = tg.config["glogin.client_secrets_file"]
            if not os.path.isfile(client_secrets_file):
                return None
            self._flow = gflow.Flow.from_client_secrets_file(
                client_secrets_file,
                scopes=[
                    "openid",
                    "https://www.googleapis.com/auth/userinfo.email",
                    "https://www.googleapis.com/auth/userinfo.profile",
                ],
            )
        return self._flow

    def identify(self, environ):
        if not self.flow:
            return None

        request = Request(environ)
        self.flow.redirect_uri = f"{request.application_url}/post_login"
        try:
            code = request.GET["code"]
        except KeyError:
            return None

        return {"code": code, "identifier": "glogin"}

    def authenticate(self, environ, identity):
        if not self.flow:
            return None

        try:
            code = identity["code"]
        except KeyError:
            return None

        if identity["identifier"] != "glogin":
            return None

        self.flow.fetch_token(code=code)
        with gdiscovery.build(
            "oauth2", "v2", credentials=self.flow.credentials
        ) as service:
            userinfo = service.userinfo().get().execute()

        email = userinfo["email"]
        user = DBSession.query(User).filter(User.email == email).one_or_none()
        if not user:
            # TODO: Handle unregistered users?
            return None

        return user.username

    def challenge(self, environ, status, app_headers, forget_headers):
        """
        Provide ``IChallenger`` interface.
        """
        if not self.flow:
            return None

        challenger = environ.get("repoze.who.challenge")
        if challenger != "glogin":
            return None
        request = Request(environ)
        self.flow.redirect_uri = f"{request.application_url}/post_login"
        auth_url, state = self.flow.authorization_url()
        headers = [
            ("Location", auth_url),
            *forget_headers,
            *((h, v) for h, v in app_headers if h.lower() == "set-cookie"),
        ]
        return HTTPFound(headers=headers)


class AuthMetadata(TGAuthMetadata):
    def __init__(self, sa_auth, *args, **kwargs):
        self.sa_auth = sa_auth
        super().__init__(*args, **kwargs)

    def get_user(self, identity, userid):
        return DBSession.query(User).filter_by(username=userid).first()

    def get_groups(self, identity, userid):
        return ["admin"] if identity["user"].admin else []

    def get_permissions(self, identity, userid):
        return ["admin"] if identity["user"].admin else []



import dataclasses
import os
import sys
from pathlib import Path

import click
import toml

import algobowl.lib.problem as problemlib
from algobowl.cli import formatter


def get_default_config_path():
    xdg_config_home = os.getenv("XDG_CONFIG_HOME")
    if not xdg_config_home:
        xdg_config_home = Path.home() / ".config"
    return Path(xdg_config_home) / "algobowl" / "cli.toml"


def _ensure_protocol(server):
    if server.startswith("http://") or server.startswith("https://"):
        return server
    if server.endswith(":8080"):
        return f"http://{server}"
    return f"https://{server}"


@dataclasses.dataclass
class CLIConfig:
    path: Path
    server: str = ""
    config: dict = dataclasses.field(default_factory=dict)
    orig_config: dict = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        if not self.path.is_file():
            return
        with open(self.path) as f:
            self.config = toml.load(f)

    def write(self):
        if self.config != self.orig_config:
            self.path.parent.mkdir(exist_ok=True, parents=True)
            with open(self.path, "w") as f:
                toml.dump(self.config, f)

    def get_server(self):
        if self.server:
            return _ensure_protocol(self.server)
        default = self.config.get("defaults", {}).get("server")
        if default:
            return _ensure_protocol(default)
        return "https://mines.algobowl.org"

    def get_url(self, path):
        return f"{self.get_server()}{path}"

    def set_default_server(self, server):
        defaults = self.config.get("defaults", {})
        defaults["server"] = server
        self.config["defaults"] = defaults

    def get_server_config(self, key, default=None):
        server = self.get_server()
        server_config = self.config.get(server, {})
        return server_config.get(key, default)

    def set_server_config(self, key, value):
        server = self.get_server()
        self.config.setdefault(server, {})
        self.config[server][key] = value

    def load_problem(self, name):
        def _propose_paths():
            path = Path(name)
            yield path
            if path.name == "problem.py":
                yield path.parent
            defaults = self.config.get("defaults", {})
            for search_path in defaults.get("problem_search_paths", []):
                yield Path(search_path) / name

        for path in _propose_paths():
            if path.is_dir() and (path / "problem.py").is_file():
                return problemlib.Problem(path)

        raise FileNotFoundError(f"Unable to find problem {name!r}")


@click.group(help="Get and set CLI configuration")
def config():
    pass


@config.command()
@click.pass_obj
def show(cli):
    cli.formatter.dump_config(cli.config.config)


@config.command()
@click.argument("server_url")
@click.pass_obj
def set_default_server(cli, server_url):
    cli.config.set_default_server(server_url)


@config.command()
@click.argument("path", type=Path)
@click.pass_obj
def add_problem_search_path(cli, path):
    if not path.is_dir():
        formatter.err(f"Path must be a directory: {path}")
        sys.exit(1)
    defaults = cli.config.config.setdefault("defaults", {})
    paths = defaults.setdefault("problem_search_paths", [])
    paths.append(str(path.resolve()))



"""
Template Helpers for AlgoBOWL
-----------------------------

This module is exposed in all templates under the name ``h``.
"""

from datetime import date, datetime, time, timedelta

import tg


def ftime(dt, duration=None, show_day=False):
    """
    Format a ``date``, ``datetime``, or ``time`` object for view on
    the page. Optionally takes a duration to show a range of time.

    :param dt: A ``date``, ``datetime`` or ``time`` to format.
    :param duration: A ``timedelta`` indicating how long something
                     lasts.
    :param show_day: Show the day of the week.
    :rtype: string
    """
    if show_day:
        date_fmt = tg.config.get("locale.dow_date_fmt", "%A, %B %-d, %Y")
    else:
        date_fmt = tg.config.get("locale.date_fmt", "%B %-d, %Y")
    time_fmt = tg.config.get("locale.time_fmt", "%-I:%M %p")
    dt_sep = tg.config.get("locale.dt_sep", " at ")
    dt_duration_sep = tg.config.get("locale.dt_duration_sep", " from ")
    duration_time_sep = tg.config.get("locale.duration_time_sep", " to ")

    if isinstance(duration, timedelta):
        # convert a duration to an end date
        duration = dt + duration

    if duration is not None and duration < dt:
        raise ValueError("Duration cannot be before the specified time")

    if isinstance(dt, datetime):
        if duration is None:
            # Format date without duration
            return dt.strftime(date_fmt + dt_sep + time_fmt)
        elif isinstance(duration, datetime):
            # Format date with duration
            start = (
                dt.strftime(date_fmt + dt_duration_sep + time_fmt) + duration_time_sep
            )
            if dt.date() == duration.date():
                return start + duration.strftime(time_fmt)
            return start + duration.strftime(date_fmt + dt_sep + time_fmt)
        else:
            raise TypeError("duration must be a timedelta or datetime")

    if isinstance(dt, date):
        return dt.strftime(date_fmt)

    if isinstance(dt, time):
        return dt.strftime(time_fmt)



import pathlib
import re
import sys

import click

import algobowl.cli.auth as auth
import algobowl.cli.formatter as fmt


def get_active_groups(cli, params=None):
    r = cli.session.get(cli.config.get_url("/group.json"), params=params)
    auth.check_response(r)
    data = r.json()
    return data["groups"]


def get_group_id(cli):
    if cli.selected_group > 0:
        return cli.selected_group
    active_groups = get_active_groups(cli)
    if len(active_groups) == 1:
        return active_groups[0]["id"]
    if not active_groups:
        fmt.err("You aren't a member of any groups.")
    else:
        fmt.err("You're a member of multiple groups.")
        fmt.err("Pass --group-id to manually select a group.")
        fmt.err(
            f"Hint: Run {fmt.cmd('algobowl group list')} to show your active groups."
        )
    sys.exit(1)


@click.group(help="Your Group")
@click.option("--group-id", type=int, help="Select group")
@click.pass_obj
def group(cli, group_id):
    if group_id:
        cli.selected_group = group_id


@group.command(name="list", help="List groups")
@click.option("--inactive/--no-inactive", help="Include inactive groups")
@click.option(
    "--non-member/--no-non-member",
    help="(Admin only) Include groups you're not a member of",
)
@click.option("--competition", type=int, help="Only show groups for this competition")
@click.pass_obj
def list_(cli, inactive, non_member, competition):
    params = {}
    if inactive:
        params["list_inactive"] = True
    if non_member:
        params["list_non_member"] = True
    if competition:
        params["competition_id"] = competition
    groups = get_active_groups(cli, params=params)
    cli.formatter.dump_table(
        groups,
        filter_keys=["id", "name", "competition_id", "incognito"],
    )


@group.command(help="Set your team name")
@click.argument("team_name")
@click.pass_obj
def set_team_name(cli, team_name):
    group_id = get_group_id(cli)
    r = cli.session.post(
        cli.config.get_url(f"/group/{group_id}/input_upload"),
        data={
            "team_name": team_name,
        },
    )
    auth.check_response(r)


@group.group(help="Upload and see your group's input")
def input():
    pass


@input.command(help="Upload your group's input")
@click.argument("file_path", type=pathlib.Path)
@click.pass_obj
def upload(cli, file_path):
    group_id = get_group_id(cli)
    with open(file_path, "rb") as f:
        r = cli.session.post(
            cli.config.get_url(f"/group/{group_id}/input_upload_api"),
            files={"input_upload": f},
        )
    auth.check_response(r)
    result = r.json()
    if result["status"] != "success":
        fmt.err(result["msg"])
        sys.exit(1)


@input.command(help="Download your group's input")
@click.argument("output_file", type=click.File("w"), default=sys.stdout)
@click.pass_obj
def download(cli, output_file):
    group_id = get_group_id(cli)
    r = cli.session.get(cli.config.get_url(f"/files/input_group{group_id}.txt"))
    auth.check_response(r)
    output_file.write(r.text)


@group.group(help="Upload and view outputs")
@click.option(
    "--to-group-id",
    type=int,
    help=(
        "Manually specify the group to submit to (by default, infer from the file name)"
    ),
)
@click.pass_obj
def output(cli, to_group_id):
    cli.to_group_id = to_group_id


def infer_group_from_path(file_path):
    matches = re.findall(r"\d+", file_path.name)
    if not matches:
        fmt.err(
            "Unable to infer the group id from file name.  "
            "Pass --to-group-id to specify manually."
        )
        sys.exit(1)
    return int(matches[-1])


@output.command(help="Upload output")
@click.argument("file_path", type=pathlib.Path)
@click.pass_obj
def upload(cli, file_path):
    from_group_id = get_group_id(cli)
    to_group_id = cli.to_group_id
    if not to_group_id:
        to_group_id = infer_group_from_path(file_path)
    with open(file_path, "rb") as f:
        r = cli.session.post(
            cli.config.get_url(f"/group/{from_group_id}/submit_output"),
            data={"to_group": to_group_id},
            files={"output_file": f},
        )
    auth.check_response(r)
    result = r.json()
    if result["status"] != "success":
        fmt.err(result["msg"])
        sys.exit(1)


@output.command(help="Download output")
@click.argument("file_path", type=pathlib.Path)
@click.pass_obj
def download(cli, file_path):
    from_group_id = get_group_id(cli)
    to_group_id = cli.to_group_id
    if not to_group_id:
        to_group_id = infer_group_from_path(file_path)
    r = cli.session.get(
        cli.config.get_url(f"/files/output_from_{from_group_id}_to_{to_group_id}.txt"),
    )
    auth.check_response(r)
    file_path.write_text(r.text)


@output.command(name="list", help="List outputs this group (will) provide")
@click.pass_obj
def output_list(cli):
    group_id = get_group_id(cli)
    r = cli.session.get(
        cli.config.get_url(f"/group/{group_id}/output_upload_list_api.json"),
    )
    auth.check_response(r)
    cli.formatter.dump_table(r.json()["outputs"])


@group.group(help="Preform verifications")
@click.pass_obj
def verification(cli):
    pass


@verification.command(name="list", help="Show verifications")
@click.pass_obj
def verification_list(cli):
    group_id = get_group_id(cli)
    r = cli.session.get(
        cli.config.get_url(f"/group/{group_id}/verification_data_v2.json"),
    )
    auth.check_response(r)
    cli.formatter.dump_table(r.json()["verifications"])



"""
Default configuration for AlgoBOWL
----------------------------------

Configuration in this file is overridden by the paste config.
"""

import tg
from depot.manager import DepotManager
from markupsafe import Markup
from repoze.who.interfaces import IChallenger
from tg.configuration import AppConfig, milestones
from tg.support.converters import asbool
from tgext.admin.config import AdminConfig
from tgext.admin.layouts import BootstrapAdminLayout
from tgext.admin.widgets import BootstrapAdminTableFiller

import algobowl
from algobowl import model
from algobowl.config.auth import AuthMetadata, GoogleAuth, MPAPIAuthenticator, TokenAuth

base_config = AppConfig()
base_config.renderers = []
base_config["flash.default_status"] = "primary"

# True to prevent dispatcher from striping extensions
# For example /socket.io would be served by "socket_io"
# method instead of "socket".
base_config.disable_request_extensions = False

# Set None to disable escaping punctuation characters to "_"
# when dispatching methods.
# Set to a function to provide custom escaping.
base_config.dispatch_path_translator = True

base_config.use_toscawidgets = False
base_config.use_toscawidgets2 = True

base_config.package = algobowl

base_config.renderers.append("json")
base_config.renderers.append("kajiki")
base_config["templating.kajiki.strip_text"] = False
base_config.default_renderer = "kajiki"

base_config["site.branding.name"] = "AlgoBOWL"
base_config["site.edu"] = True

# Configure Sessions, store data as JSON to avoid pickle security issues
base_config["session.enabled"] = True
base_config["session.data_serializer"] = "json"

# Configure the base SQLALchemy Setup
base_config.use_sqlalchemy = True
base_config.model = model
base_config.DBSession = model.DBSession

# Configure the authentication backend
base_config.auth_backend = "sqlalchemy"
base_config.sa_auth.cookie_secret = "aa8deb7d-a235-4f8d-807a-6497a8bb26af"
base_config.sa_auth.authmetadata = AuthMetadata(base_config.sa_auth)
base_config["identity.allow_missing_user"] = False

authenticators = [
    ("token", TokenAuth()),
    ("mpapi", MPAPIAuthenticator()),
    ("glogin", GoogleAuth()),
]

base_config.sa_auth.authenticators = authenticators
base_config.sa_auth.identifiers = [
    ("default", None),
    *authenticators,
]
base_config.sa_auth.challengers = [
    auth for auth in authenticators if IChallenger.providedBy(auth[1])
]

base_config["auth.mpapi.url"] = "https://mastergo.mines.edu/mpapi"


def variable_provider():
    return {"asbool": asbool}


base_config.variable_provider = variable_provider


def config_ready():
    """Executed once the configuration is ready."""
    # Configure default depot
    depot_config = dict(tg.config)
    if "depot.backend" not in depot_config:
        depot_config.setdefault("depot.storage_path", "/tmp/depot")
    DepotManager.configure("default", depot_config)


milestones.config_ready.register(config_ready)


class AdminTableFiller(BootstrapAdminTableFiller):
    def __actions__(self, obj):
        primary_fields = self.__provider__.get_primary_fields(self.__entity__)
        pklist = "/".join(map(lambda x: str(getattr(obj, x)), primary_fields))

        return Markup(
            """
    <a href="{pklist}/edit" class="btn btn-secondary">
        <i class="fas fa-pen fa-fw"></i>
    </a>
    <form method="POST" action="{pklist}" style="display: inline">
        <input type="hidden" name="_method" value="DELETE" />
        <button type="submit" class="btn btn-danger"
                onclick="return confirm('Are you sure?')">
            <i class="fas fa-trash fa-fw"></i>
        </button>
    </form>""".format(
                pklist=pklist
            )
        )


class AdminLayout(BootstrapAdminLayout):
    template_index = "algobowl.templates.admin.index"
    crud_templates = {
        "get_all": ["kajiki:algobowl.templates.admin.get_all"],
        "edit": ["kajiki:algobowl.templates.admin.edit"],
        "new": ["kajiki:algobowl.templates.admin.new"],
    }
    TableFiller = AdminTableFiller


class AdminConfig(AdminConfig):
    project_name = "AlgoBOWL"
    layout = AdminLayout
    allow_only = tg.predicates.has_permission("admin")


try:
    # Enable DebugBar if available, install tgext.debugbar to turn it on
    from tgext.debugbar import enable_debugbar

    enable_debugbar(base_config)
except ImportError:
    pass



import datetime

import sqlalchemy
import tg

import algobowl.lib.base as base
import algobowl.model as model


class PrefController(base.BaseController):
    allow_only = tg.predicates.not_anonymous()

    @tg.expose("algobowl.templates.cli_setup")
    @tg.expose("json")
    def cli(self, client_name=None, client_id=None):
        user = tg.request.identity["user"]

        if client_name and client_id:
            if len(client_id) != 88:
                tg.abort(400, "client_id must be 88 characters long")
            if " " in client_id:
                tg.abort(400, "client_id cannot contain spaces")
            token = model.AuthToken(
                client_name=client_name,
                client_id=client_id,
                date_added=datetime.datetime.now(),
                user=user,
            )
            try:
                model.DBSession.add(token)
                model.DBSession.flush()
            except sqlalchemy.exc.IntegrityError:
                tg.flash("Cannot add client (already added?)", "danger")
                model.DBSession.rollback()
            else:
                tg.flash(f"Client {client_name} successfully added!", "success")
            tg.redirect(tg.url("/pref/cli"))

        identity = tg.request.environ.get("repoze.who.identity", {})
        this_client_id = identity.get("token")
        auth_tokens_json = [
            {
                "id": token.id,
                "client_name": token.client_name,
                "date_added": token.date_added.isoformat(),
                "this_client": this_client_id == token.client_id,
            }
            for token in user.auth_tokens
        ]

        return dict(auth_tokens=auth_tokens_json)

    @tg.expose()
    def revoke_auth_token(self, token_id):
        try:
            token_id = int(token_id)
        except ValueError:
            tg.abort(400, "token_id must be an integer")

        user = tg.request.identity["user"]
        query = model.DBSession.query(model.AuthToken).filter(
            model.AuthToken.id == token_id
        )
        token = query.one_or_none()

        if not user.admin and token.user_id != user.id:
            tg.abort(403, "You don't have permission to revoke this token")

        client_name = token.client_name
        query.delete()
        model.DBSession.flush()
        tg.flash(f"Client {client_name} revoked!", "success")
        tg.redirect(tg.url("/pref/cli"))

    @tg.expose("json")
    def whoami(self):
        user = tg.request.identity["user"]
        return {
            "user_id": user.id,
            "username": user.username,
            "full_name": user.full_name,
            "email": user.email,
        }



import contextlib
import functools
import io

from tg import expose

__all__ = ["logoutput"]


def logoutput(meth):
    """Decorator to use log output"""

    @functools.wraps(meth)
    def wrapper(*args, **kwargs):
        buf = io.StringIO()
        with contextlib.redirect_stdout(buf):
            meth(*args, **kwargs)
        return {"content": buf.getvalue()}

    return expose("algobowl.templates.logoutput")(wrapper)



import concurrent.futures
import io
import random
import sys
from pathlib import Path

import click

import algobowl.lib.problem as problemlib
from algobowl.cli import formatter


def _parse_input(cli, path: Path):
    with open(path, "r", encoding="ascii") as f:
        try:
            return cli.problem.parse_input(f)
        except problemlib.FileFormatError as e:
            formatter.err(f"Input has formatting errors: {e}")
            sys.exit(1)


def _parse_output(cli, iput, oput_path: Path):
    if isinstance(iput, Path):
        iput = _parse_input(cli, iput)
    with open(oput_path, "r", encoding="ascii") as f:
        try:
            return cli.problem.parse_output(iput, f)
        except problemlib.FileFormatError as e:
            formatter.err(f"Output has formatting errors: {e}")
            sys.exit(1)


@click.group(help="Work with local problems")
@click.argument("problem", type=str)
@click.pass_obj
def problem(cli, problem):
    cli.problem = cli.config.load_problem(problem)


@problem.command(help="Run tests on a problem")
@click.argument("pytest_args", nargs=-1)
@click.pass_obj
def test(cli, pytest_args):
    pytest_args = [
        "--cov",
        "problem",
        "--cov-report",
        "term-missing:skip-covered",
        *pytest_args,
    ]
    try:
        cli.problem.run_tests(pytest_args)
    except problemlib.ProblemTestError:
        sys.exit(1)


@problem.command(help="Parse an input file and print the normalized version")
@click.argument("input_file", type=Path)
@click.pass_obj
def parse_input(cli, input_file):
    iput = _parse_input(cli, input_file)
    iput.write(sys.stdout)


@problem.command(help="Parse an output file and print the normalized version")
@click.argument("input_file", type=Path)
@click.argument("output_file", type=Path)
@click.pass_obj
def parse_output(cli, input_file, output_file):
    oput = _parse_output(cli, input_file, output_file)
    oput.write(sys.stdout)


@problem.command(help="Verify an output")
@click.argument("input_file", type=Path)
@click.argument("output_files", type=Path, nargs=-1, required=True)
@click.pass_obj
def verify(cli, input_file, output_files):
    rv = 0
    iput = _parse_input(cli, input_file)

    def _verify_output(output_file):
        rv = 0
        oput = _parse_output(cli, iput, output_file)
        message = click.style("OK!", fg="green", bold=True)
        try:
            oput.verify()
        except problemlib.VerificationError as e:
            rv = 1
            message = click.style(f"BAD: {e}", fg="red", bold=True)
        message = f"{output_file}: {message}"
        return rv, message

    with concurrent.futures.ThreadPoolExecutor() as executor:
        futures = []
        for output_file in output_files:
            futures.append(executor.submit(_verify_output, output_file))

    for future in futures:
        future_rv, message = future.result()
        rv = rv or future_rv
        click.echo(message)

    sys.exit(rv)


@problem.command(help="Generate an input")
@click.option("--seed", type=int, help="Seed to use.  Default: use SystemRandom.")
@click.pass_obj
def generate_input(cli, seed):
    if seed:
        rng = random.Random(seed)
    else:
        rng = random.SystemRandom()
    iput = cli.problem.generate_input(rng)
    iput.write(sys.stdout)


@problem.command(help="Trivially solve an input")
@click.argument("input_file", type=Path)
@click.pass_obj
def trivial_solve(cli, input_file):
    iput = _parse_input(cli, input_file)
    output = iput.trivial_solve()
    output.write(sys.stdout)



import json

import click
import tabulate
import toml


def cmd(text):
    return click.style(text, fg="cyan", bold=True)


def err(text):
    err_prefix = click.style("ERROR:", fg="red", bold=True)
    click.echo(f"{err_prefix} {text}", err=True)


def filter_table(table, keys):
    table = [{k: row[k] for k in keys} for row in table]
    return table


class JsonFormatter:
    def dump_table(self, table, headers="keys", filter_keys=None):
        if filter_keys:
            table = filter_table(table, filter_keys)
        self.dump_json(table)

    def dump_config(self, config):
        self.dump_json(config)

    def dump_json(self, data):
        click.echo(json.dumps(data, sort_keys=True, indent=4))


class TableFormatter(JsonFormatter):
    def dump_table(self, table, headers="keys", filter_keys=None):
        if filter_keys:
            table = filter_table(table, filter_keys)
        click.echo(tabulate.tabulate(table, headers=headers))

    def dump_config(self, config):
        click.echo(toml.dumps(config), nl=False)



import dataclasses

import tg

from algobowl import model
from algobowl.lib import base


@dataclasses.dataclass
class UserApiController(base.BaseController):
    user: model.User

    @tg.expose("json")
    def info(self):
        return {
            "username": self.user.username,
            "email": self.user.email,
            "id": self.user.id,
            "full_name": self.user.full_name,
            "admin": self.user.admin,
            "groups": [group.id for group in self.user.groups],
        }


class UserLookupApiController(base.BaseController):
    allow_only = tg.predicates.not_anonymous()

    @tg.expose()
    def _lookup(self, lookup, *args):
        def _lookup_by_uid():
            try:
                user_id = int(lookup)
            except ValueError:
                return None
            return (
                model.DBSession.query(model.User)
                .filter(model.User.id == user_id)
                .one_or_none()
            )

        def _lookup_by_username():
            return model.User.from_username(lookup)

        def _lookup_by_email():
            if "@" not in lookup:
                return None
            return (
                model.DBSession.query(model.User)
                .filter(model.User.email == lookup)
                .one_or_none()
            )

        for f in (_lookup_by_uid, _lookup_by_username, _lookup_by_email):
            user = f()
            if user:
                break
        else:
            tg.abort(404, "No such user found")

        return UserApiController(user), args



import click

import algobowl.cli.auth as auth


def lookup_uid(cli, lookup):
    try:
        user_id = int(lookup)
    except ValueError:
        pass
    else:
        return user_id

    r = cli.session.get(cli.config.get_url(f"/api/user/{lookup}/info.json"))
    auth.check_response(r)
    data = r.json()
    return data["id"]


@click.group(help="Administration")
@click.pass_obj
def admin(cli):
    pass


@admin.command(name="create-group", help="Create a new group")
@click.option(
    "--incognito/--no-incognito", "-i", help="Enable if the group should be incognito"
)
@click.option("--name", "-n", type=str, help="Group Name")
@click.option(
    "--competition", "-c", type=int, help="Competition ID for this group", required=True
)
@click.option("--user", "-u", type=str, help="Add user", multiple=True)
@click.pass_obj
def create_group(cli, incognito, name, competition, user):
    user_ids = []
    for u in user:
        user_ids.append(str(lookup_uid(cli, u)))

    data = {"competition": str(competition), "users": user_ids}
    if incognito:
        data["incognito"] = "true"
    if name:
        data["name"] = name

    r = cli.session.post(cli.config.get_url("/admin/groups"), data=data)
    auth.check_response(r)



"""The application's Globals object"""

__all__ = ["Globals"]


class Globals:
    """
    Container for objects available throughout the life of the application.

    One instance of Globals is created during application initialization and
    is available during requests via the ``app_globals`` variable.
    """

    def __init__(self):
        pass



# -*- coding: utf-8 -*-
"""The base Controller API."""

from tg import TGController, request, tmpl_context

__all__ = ["BaseController"]


class BaseController(TGController):
    """
    Base class for the controllers in the application.

    Your web application should have one of these. The root of
    your application is used to compute URLs used by your app.

    """

    def __call__(self, environ, context):
        """Invoke the Controller"""
        # TGController.__call__ dispatches to the Controller method
        # the request is routed to.

        tmpl_context.identity = request.identity

        return TGController.__call__(self, environ, context)



import datetime
import enum

import sqlalchemy as sa
import tg
from depot.fields.sqlalchemy import UploadedFileField
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relation, relationship, scoped_session, sessionmaker
from zope.sqlalchemy import register

maker = sessionmaker(autoflush=True, autocommit=False)
DBSession = scoped_session(maker)
register(DBSession)
DeclarativeBase = declarative_base()
DeclarativeBase.query = DBSession.query_property()
metadata = DeclarativeBase.metadata


def init_model(engine):
    """
    Call me before using any of the tables or classes in the model.
    """
    DBSession.configure(bind=engine)
    return DBSession


class VerificationStatus(enum.Enum):
    waiting = 0
    accepted = 1
    rejected = 2

    def __str__(self):
        return self.name


class Competition(DeclarativeBase):
    __tablename__ = "competition"
    db_icon = "fas fa-clipboard-list"

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.String, nullable=False)

    # TODO(jrosenth): Make this non-nullable when all problems have
    # been migrated to the new format.
    problem = sa.Column(sa.String, nullable=True)

    allow_custom_team_names = sa.Column(sa.Boolean, default=True)

    input_upload_begins = sa.Column(sa.DateTime, nullable=False)
    input_upload_ends = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("input_upload_ends > input_upload_begins"),
        nullable=False,
    )
    output_upload_begins = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("output_upload_begins >= input_upload_ends"),
        nullable=False,
    )
    output_upload_ends = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("output_upload_ends > output_upload_begins"),
        nullable=False,
    )
    verification_begins = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("verification_begins >= output_upload_ends"),
        nullable=True,
    )
    verification_ends = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("verification_ends > verification_begins"),
        nullable=True,
    )
    resolution_begins = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("resolution_begins >= verification_ends"),
        nullable=True,
    )
    resolution_ends = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("resolution_ends > resolution_begins"),
        nullable=True,
    )
    open_verification_begins = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("open_verification_begins >= resolution_ends"),
        nullable=True,
    )
    open_verification_ends = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("open_verification_ends >= open_verification_begins"),
        nullable=True,
    )
    evaluation_begins = sa.Column(sa.DateTime, nullable=True)
    evaluation_ends = sa.Column(
        sa.DateTime,
        sa.CheckConstraint("evaluation_ends > evaluation_begins"),
        nullable=True,
    )

    groups = relationship(
        "Group",
        back_populates="competition",
        lazy="dynamic",
        order_by="Group.id",
        uselist=True,
    )

    @property
    def input_upload_open(self):
        return (
            self.input_upload_begins <= datetime.datetime.now() < self.input_upload_ends
        )

    @property
    def output_upload_open(self):
        return (
            self.output_upload_begins
            <= datetime.datetime.now()
            < self.output_upload_ends
        )

    @property
    def verification_open(self):
        return self.verification_begins and (
            self.verification_begins <= datetime.datetime.now() < self.verification_ends
        )

    @property
    def resolution_open(self):
        return self.resolution_begins and (
            self.resolution_begins <= datetime.datetime.now() < self.resolution_ends
        )

    @property
    def open_verification_open(self):
        return self.open_verification_begins and (
            self.open_verification_begins
            <= datetime.datetime.now()
            < self.open_verification_ends
        )

    @property
    def evaluation_open(self):
        return self.evaluation_begins and (
            self.evaluation_begins <= datetime.datetime.now() < self.evaluation_ends
        )

    @property
    def end(self):
        return max(
            filter(
                lambda x: x,
                (
                    self.output_upload_ends,
                    self.verification_ends,
                    self.resolution_ends,
                    self.open_verification_ends,
                    self.evaluation_ends,
                ),
            )
        )

    @property
    def published(self):
        return self.input_upload_begins <= datetime.datetime.now()

    @property
    def active(self):
        return self.input_upload_begins <= datetime.datetime.now() < self.end

    @property
    def archived(self):
        # 3 months after end
        archive_date = self.end + datetime.timedelta(days=90)
        return datetime.datetime.now() >= archive_date

    def __repr__(self):
        return self.name


class User(DeclarativeBase):
    __tablename__ = "user"
    db_icon = "fas fa-user"

    id = sa.Column(sa.Integer, primary_key=True)
    username = sa.Column(sa.Unicode, unique=True, nullable=False)
    email = sa.Column(sa.Unicode, unique=True, nullable=False)
    full_name = sa.Column(sa.Unicode)
    admin = sa.Column(sa.Boolean, nullable=False, default=False)

    submitted_evaluations = relationship(
        "Evaluation",
        back_populates="from_student",
        foreign_keys="Evaluation.from_student_id",
        lazy="dynamic",
    )
    received_evaluations = relationship(
        "Evaluation",
        back_populates="to_student",
        foreign_keys="Evaluation.to_student_id",
    )

    groups = relation("Group", secondary="user_group_xref", back_populates="users")
    auth_tokens = relationship(
        "AuthToken",
        back_populates="user",
        lazy="dynamic",
        order_by="AuthToken.date_added",
    )

    @classmethod
    def from_username(cls, username):
        return DBSession.query(User).filter(cls.username == username).one_or_none()

    def __repr__(self):
        return self.full_name or self.username


class AuthToken(DeclarativeBase):
    __tablename__ = "auth_token"
    db_icon = "fas fa-terminal"

    id = sa.Column(sa.Integer, primary_key=True)
    client_name = sa.Column(sa.Unicode, nullable=True)
    client_id = sa.Column(sa.Unicode, nullable=False, unique=True)
    date_added = sa.Column(sa.DateTime, nullable=False)

    user_id = sa.Column(sa.Integer, sa.ForeignKey("user.id"), nullable=False)
    user = relationship("User", back_populates="auth_tokens")


class Group(DeclarativeBase):
    __tablename__ = "group"
    db_icon = "fas fa-users"

    id = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.Unicode(100), nullable=True)
    incognito = sa.Column(sa.Boolean, default=False)
    benchmark = sa.Column(sa.Boolean, default=False)

    competition_id = sa.Column(
        sa.Integer, sa.ForeignKey("competition.id"), nullable=False
    )
    competition = relationship("Competition", back_populates="groups")

    input = relationship("Input", uselist=False, back_populates="group")
    outputs = relationship(
        "Output", back_populates="group", lazy="dynamic", order_by="Output.group_id"
    )

    users = relation(
        "User", secondary="user_group_xref", back_populates="groups", lazy="dynamic"
    )

    protests = relationship("Protest", back_populates="submitter", lazy="dynamic")

    evaluations = relationship("Evaluation", back_populates="group")

    def __repr__(self):
        if self.name:
            return "{} (Group {})".format(self.name, self.id)
        return "Group {}".format(self.id)


# N users can be a part of M groups
# ---------------------------------
# Why may a user be a part of M groups? Consider the case where a user
# fails the course, or a test account that was made part of many groups
# over time.
user_group_xref = sa.Table(
    "user_group_xref",
    metadata,
    sa.Column(
        "user_id",
        sa.Integer,
        sa.ForeignKey("user.id", onupdate="CASCADE", ondelete="CASCADE"),
        primary_key=True,
    ),
    sa.Column(
        "group_id",
        sa.Integer,
        sa.ForeignKey("group.id", onupdate="CASCADE", ondelete="CASCADE"),
        primary_key=True,
    ),
)


class Input(DeclarativeBase):
    __tablename__ = "input"
    db_icon = "far fa-file"

    id = sa.Column(sa.Integer, primary_key=True)
    data = sa.Column(UploadedFileField, nullable=False)
    is_default = sa.Column(sa.Boolean, nullable=True)

    group_id = sa.Column(sa.Integer, sa.ForeignKey("group.id"), nullable=False)
    group = relationship("Group", back_populates="input")

    outputs = relationship(
        "Output", back_populates="input", lazy="dynamic", order_by="Output.group_id"
    )

    @property
    def filename(self):
        return f"input_group{self.group_id}.txt"

    @property
    def url(self):
        return tg.url(f"/files/{self.filename}")

    def __repr__(self):
        return "Input from [{!r}]".format(self.group)


class Output(DeclarativeBase):
    __tablename__ = "output"
    db_icon = "far fa-copy"

    id = sa.Column(sa.Integer, primary_key=True)

    # The score is separated from the data, allowing for database
    # server-side sorting
    score = sa.Column(sa.BigInteger, nullable=False)
    data = sa.Column(UploadedFileField, nullable=False)

    verification = sa.Column(
        sa.Enum(VerificationStatus), nullable=False, default=VerificationStatus.waiting
    )
    ground_truth = sa.Column(
        sa.Enum(VerificationStatus), nullable=False, default=VerificationStatus.waiting
    )

    active = sa.Column(sa.Boolean, nullable=False, default=True)
    original = sa.Column(sa.Boolean, nullable=False, default=True)
    use_ground_truth = sa.Column(sa.Boolean, nullable=False, default=False)

    input_id = sa.Column(sa.Integer, sa.ForeignKey("input.id"), nullable=False)
    input = relationship("Input", back_populates="outputs")

    group_id = sa.Column(sa.Integer, sa.ForeignKey("group.id"), nullable=False)
    group = relationship("Group", back_populates="outputs")

    protests = relationship("Protest", back_populates="output", lazy="dynamic")

    @property
    def filename(self):
        return f"output_from_{self.group_id}_to_{self.input.group_id}.txt"

    @property
    def url(self):
        return tg.url(f"/files/{self.filename}")

    def __repr__(self):
        return "Output from [{!r}] for {!r}".format(self.group, self.input)


class Protest(DeclarativeBase):
    """
    A protest submitted during output verification. The same term
    (protest) is overloaded for the resolution stage and *does not*
    create one of these. In the resolution stage, the protest feature
    simply reveals the ground truth and disables reupload.
    """

    __tablename__ = "protest"
    db_icon = "far fa-thumbs-down"

    id = sa.Column(sa.Integer, primary_key=True)
    message = sa.Column(sa.Unicode(1000), nullable=False)
    accepted = sa.Column(sa.Boolean, nullable=False)

    submitter_id = sa.Column(sa.Integer, sa.ForeignKey("group.id"), nullable=False)
    submitter = relationship("Group", back_populates="protests")

    output_id = sa.Column(sa.Integer, sa.ForeignKey("output.id"), nullable=False)
    output = relationship("Output", back_populates="protests")

    def __repr__(self):
        return "Protest {}, from [{!r}] on {!r}".format(
            self.id, self.submitter, self.output
        )


class Evaluation(DeclarativeBase):
    __tablename__ = "evaluation"
    db_icon = "fas fa-sliders-h"

    id = sa.Column(sa.Integer, primary_key=True)
    score = sa.Column(sa.Float, nullable=False)

    from_student_id = sa.Column(sa.Integer, sa.ForeignKey("user.id"), nullable=False)
    from_student = relationship(
        "User", back_populates="submitted_evaluations", foreign_keys=from_student_id
    )

    to_student_id = sa.Column(sa.Integer, sa.ForeignKey("user.id"), nullable=False)
    to_student = relationship(
        "User", back_populates="received_evaluations", foreign_keys=to_student_id
    )

    group_id = sa.Column(sa.Integer, sa.ForeignKey("group.id"), nullable=False)
    group = relationship("Group", back_populates="evaluations")



import datetime
import re
import zipfile
from io import BytesIO, StringIO

from depot.io.utils import FileIntent
from tg import abort, expose, flash, redirect, request, require, response, url
from tg.predicates import has_permission, not_anonymous

import algobowl.lib.problem as problemlib
from algobowl.lib.base import BaseController
from algobowl.lib.helpers import ftime
from algobowl.model import (
    DBSession,
    Evaluation,
    Group,
    Input,
    Output,
    VerificationStatus,
)

__all__ = ["GroupsController", "GroupController"]

newline_p = re.compile(rb"\s*\n")
spacesep_p = re.compile(rb"[ \t\v\f]+")


class GroupController(BaseController):
    allow_only = not_anonymous()

    def __init__(self, group):
        self.group = group
        self.base_url = "/group/{}".format(group.id)

    @expose("algobowl.templates.group.index")
    def index(self):
        stage = None
        if self.group.competition.input_upload_open:
            stage = "input_upload"
        if self.group.competition.output_upload_open:
            stage = "output_upload"
        if self.group.competition.verification_open:
            stage = "verification"
        if self.group.competition.resolution_open:
            stage = "resolution"
        if self.group.competition.evaluation_open:
            stage = "evaluation"

        return self.stagepage(stage)

    @expose("algobowl.templates.group.index")
    def stage(self, stage):
        user = request.identity["user"]
        if not user.admin:
            abort(403, "Manual stage selection is for admin users only.")
        return self.stagepage(stage)

    def stagepage(self, stage):
        user = request.identity["user"]
        submitted_outputs = {o.input.id: o for o in self.group.outputs}
        d = {
            "competition": self.group.competition,
            "group": self.group,
            "submitted_outputs": submitted_outputs,
            "stage": stage,
        }
        if stage == "evaluation":
            evals = dict(
                DBSession.query(Evaluation.to_student_id, Evaluation.score)
                .filter(Evaluation.group_id == self.group.id)
                .filter(Evaluation.from_student_id == user.id)
            )
            for member in self.group.users:
                if member.id not in evals.keys():
                    evals[member.id] = 1.0
            d["evals"] = evals
        return d

    @expose("json")
    def output_upload_list_api(self):
        outputs = {}
        if datetime.datetime.now() < self.group.competition.output_upload_begins:
            return {"outputs": []}
        for group in self.group.competition.groups:
            if group.input is not None:
                outputs[group.id] = {
                    "input_group_id": group.id,
                    "submitted": False,
                }
        for output in self.group.outputs:
            outputs[output.input.group.id].update(
                {
                    "submitted": True,
                    "score": output.score,
                    "url": output.url,
                }
            )
        return {
            "outputs": [value for key, value in sorted(outputs.items())],
        }

    @expose("json")
    def input_upload_api(self, input_upload):
        try:
            contents = input_upload.file.read().decode("utf-8")
        except UnicodeDecodeError:
            return {
                "status": "error",
                "msg": (
                    "Your input contains invalid characters.  Please correct and try "
                    "uploading again."
                ),
            }

        problem = problemlib.load_problem(self.group.competition)
        try:
            input = problem.parse_input(StringIO(contents))
        except problemlib.FileFormatError as e:
            return {
                "status": "error",
                "msg": (
                    f"Your input is not valid: {e}.  Please correct and upload again."
                ),
            }

        reformatted_contents = StringIO()
        input.write(reformatted_contents)

        f = FileIntent(
            BytesIO(reformatted_contents.getvalue().encode("utf-8")),
            "input_group{}.txt".format(self.group.id),
            "application/octet-stream",
        )
        if self.group.input is None:
            iput = Input(data=f, group=self.group, is_default=False)
            DBSession.add(iput)
        else:
            self.group.input.data = f
            self.group.input.is_default = False
        DBSession.flush()

        return {"status": "success"}

    @expose()
    def input_upload(self, input_upload=None, team_name=None):
        user = request.identity["user"]
        if not user.admin and not self.group.competition.input_upload_open:
            abort(403, "Sorry, input upload stage is closed.")

        if hasattr(input_upload, "file"):
            result = self.input_upload_api(input_upload)
            if result["status"] != "success":
                flash(result["msg"], "danger")
                redirect(self.base_url)

        if team_name is not None:
            if len(team_name) >= 100:
                flash("Your team name has been rejected, as it is too long.", "danger")
                redirect(self.base_url)

            if not team_name:
                flash("You must set a team name.", "danger")
                redirect(self.base_url)

            self.group.name = team_name

        flash(
            "Thank you. Please return here for output upload on {}".format(
                ftime(self.group.competition.output_upload_begins)
            ),
            "success",
        )
        redirect(self.base_url)

    @expose("json")
    def submit_output(self, to_group, output_file=None):
        user = request.identity["user"]
        to_group = DBSession.query(Group).get(to_group)
        if not to_group:
            abort(404, "No such group")
        if to_group.competition_id != self.group.competition_id:
            abort(400, "Cannot submit to a group in another competition")
        if not hasattr(output_file, "file"):
            abort(400, "Must include file in submission")
        comp = self.group.competition
        existing = (
            DBSession.query(Output)
            .filter(Output.group_id == self.group.id)
            .filter(Output.input_id == to_group.input.id)
            .filter(Output.active == True)
            .one_or_none()
        )
        if not (
            user.admin
            or comp.output_upload_open
            or (
                comp.resolution_open
                and existing is not None
                and not existing.use_ground_truth
            )
            or (comp.resolution_open and existing is None)
        ):
            abort(403, "Forbidden to upload this output at this time")

        try:
            contents = output_file.file.read().decode("utf-8")
        except UnicodeDecodeError:
            return {"status": "error", "msg": "Output contains invalid characters."}

        problem = problemlib.load_problem(comp)

        try:
            output = problem.parse_output(to_group.input.data.file, StringIO(contents))
        except problemlib.FileFormatError as e:
            return {
                "status": "error",
                "msg": f"Output has formatting errors: {e}",
            }

        new_contents = StringIO()
        output.write(new_contents)

        f = FileIntent(
            BytesIO(new_contents.getvalue().encode("utf-8")),
            "output_from_{}_to_{}.txt".format(self.group.id, to_group.id),
            "application/octet-stream",
        )
        try:
            output.verify()
        except problemlib.VerificationError:
            ground_truth = VerificationStatus.rejected
        except Exception:
            ground_truth = VerificationStatus.waiting
        else:
            ground_truth = VerificationStatus.accepted

        use_ground_truth = (
            not comp.verification_begins
            or datetime.datetime.now() >= comp.verification_begins
        )

        if existing:
            if comp.resolution_open:
                existing.active = False
            else:
                DBSession.delete(existing)

        db_output = Output(
            data=f,
            group=self.group,
            input=to_group.input,
            score=output.score,
            original=comp.output_upload_open,
            ground_truth=ground_truth,
            use_ground_truth=use_ground_truth,
        )
        DBSession.add(db_output)
        DBSession.flush()

        return {"status": "success", "url": db_output.data.url}

    @expose("json")
    def submit_verification(self, output_id, status):
        user = request.identity["user"]
        try:
            output_id = int(output_id)
        except ValueError:
            return {
                "status": "error",
                "msg": "There was an error. Please try clearing your "
                "browser cache and repeating this action.",
            }
        output = DBSession.query(Output).get(output_id)
        if not output or output.input.group_id != self.group.id:
            abort(404)
        try:
            status = VerificationStatus[status]
        except KeyError:
            abort(404)
        if not user.admin and not self.group.competition.verification_open:
            return {"status": "error", "msg": "Verification closed"}
        assert output.original is True
        assert output.active is True
        assert output.use_ground_truth is False
        output.verification = status
        return self.verification_data()

    @expose("json")
    def verification_data(self):
        data = {
            k: str(v)
            for k, v in (
                DBSession.query(Output.id, Output.verification)
                .join(Output.input)
                .filter(Input.group_id == self.group.id)
            )
        }
        return {"status": "success", "data": data}

    @expose("json")
    def verification_data_v2(self):
        our_input = (
            DBSession.query(Input).filter(Input.group_id == self.group.id).one_or_none()
        )

        if not our_input:
            abort(404, "Your group has no input.")

        outputs = DBSession.query(Output).filter(Output.input_id == our_input.id)

        result = []
        for output in outputs:
            result.append(
                {
                    "input_filename": our_input.filename,
                    "output_filename": output.filename,
                    "status": str(output.verification),
                }
            )
        return {"verifications": result}

    @expose()
    def verification_outputs(self):
        user = request.identity["user"]
        if not user.admin and not self.group.competition.verification_open:
            abort(403, "This file is only available during verification.")
        f = BytesIO()
        archive = zipfile.ZipFile(f, mode="w")
        outputs = (
            DBSession.query(Output)
            .join(Output.input)
            .filter(Input.group_id == self.group.id)
        )
        for output in outputs:
            archive.writestr(
                "verification_outputs/{}".format(output.data.filename),
                output.data.file.read(),
            )
        archive.close()
        f.seek(0)
        response.content_type = "application/zip"
        return f.read()

    @expose()
    @require(has_permission("admin"))
    def automatic_verification(self):
        outputs = (
            DBSession.query(Output)
            .join(Output.input)
            .filter(Input.group_id == self.group.id)
        )
        for output in outputs:
            output.verification = output.ground_truth
            DBSession.add(output)
        DBSession.flush()

    @expose("json")
    def resolution_protest(self, output_id):
        if not self.group.competition.resolution_open:
            return {"status": "error", "msg": "Resolution stage is not open."}
        output = DBSession.query(Output).get(output_id)
        if not output or output.group_id != self.group.id:
            abort(404)
        if not output.active:
            return {
                "status": "error",
                "msg": "Output has been replaced; cannot protest.",
            }
        if output.use_ground_truth:
            return {"status": "error", "msg": "Output has already been protested."}
        assert output.original
        output.use_ground_truth = True
        return {"status": "success"}

    @expose()
    def submit_evaluations(self):
        me = request.identity["user"]
        students_allowed = set(user.id for user in self.group.users)
        for user_id, score in request.POST.items():
            user_id = int(user_id)
            if user_id not in students_allowed:
                abort(403, "You cannot submit an evaluation to this user")
            score = float(score)
            if score < 0:
                abort(403, "Negative contributions are not allowed")
            e = (
                DBSession.query(Evaluation)
                .filter(Evaluation.from_student_id == me.id)
                .filter(Evaluation.to_student_id == user_id)
                .filter(Evaluation.group_id == self.group.id)
                .one_or_none()
            )
            if not e:
                e = Evaluation(
                    from_student_id=me.id,
                    to_student_id=user_id,
                    group=self.group,
                    score=score,
                )
            else:
                e.score = score
            DBSession.add(e)
        flash("Your evaluations have been saved. Thank you.", "success")
        redirect("/group/{}".format(self.group.id))


class GroupsController(BaseController):
    allow_only = not_anonymous()

    @expose()
    def _lookup(self, group_id, *args):
        try:
            group_id = int(group_id)
        except ValueError:
            abort(404, "Not a valid group identifier.")
        user = request.identity["user"]
        group = DBSession.query(Group).get(group_id)
        if not group:
            abort(404, "No such group.")
        if user not in group.users and not user.admin:
            abort(403, "You are not a part of this group.")
        return GroupController(group), args

    @expose("algobowl.templates.group.select")
    @expose("json")
    def index(self, list_inactive=False, list_non_member=False, competition_id=None):
        user = request.identity["user"]

        all_groups = user.groups
        if list_non_member:
            if not user.admin:
                abort(403, "Only admin may use list_non_member.")
            all_groups = DBSession.query(Group).all()

        groups = []
        for group in all_groups:
            if competition_id and group.competition_id != int(competition_id):
                continue
            if list_inactive or group.competition.active:
                groups.append(group)

        if len(groups) == 1 and request.response_type != "application/json":
            redirect(url("/group/{}".format(groups[0].id)))

        return {"groups": groups}



"""Automated tester for the problem format."""

import enum
import io
import pathlib
import random
import sys

import pytest

import algobowl.lib.problem as problemlib


class ASCIIFormat(enum.Enum):
    UNIX = 0
    DOS = 1
    DOS_WITH_FINAL_TERMINATOR = 2


@pytest.fixture
def problem(problem_dir):
    return problemlib.Problem(problem_dir)


@pytest.fixture(params=[1337, 0xDEADBEEF, 0xDEADD00D, 55378008])
def rng(request):
    return random.Random(request.param)


def stringio_from_path(path, ascii_format=False):
    contents = path.read_text(encoding="ascii")
    assert "\r\n" not in contents
    if ascii_format != ASCIIFormat.UNIX:
        if contents.endswith("\n"):
            contents = contents[:-1]
        contents = contents.replace("\n", "\r\n")
        if ascii_format == ASCIIFormat.DOS_WITH_FINAL_TERMINATOR:
            contents += "\r\n"
    return io.StringIO(contents)


def test_good_input(problem, good_input_path, ascii_format):
    input = problem.parse_input(
        stringio_from_path(good_input_path, ascii_format=ascii_format)
    )
    output_buf = io.StringIO()
    input.write(output_buf)

    reformatted = output_buf.getvalue()
    assert reformatted.endswith("\n")

    # The reformatted data should also parse OK.
    problem.parse_input(io.StringIO(reformatted))


def test_bad_input(problem, bad_input_path, ascii_format):
    with pytest.raises(problemlib.FileFormatError):
        problem.parse_input(
            stringio_from_path(bad_input_path, ascii_format=ascii_format)
        )


def test_empty_input(problem):
    # The empty input should never be valid.
    with pytest.raises(problemlib.FileFormatError):
        problem.parse_input(io.StringIO(""))


def test_good_output(problem, good_output_path, ascii_format):
    input_path, output_path = good_output_path
    input = problem.parse_input(stringio_from_path(input_path))
    output = problem.parse_output(
        input, stringio_from_path(output_path, ascii_format=ascii_format)
    )
    output.verify()

    # Reformatting the output should still result in a good output.
    output_buf = io.StringIO()
    output.write(output_buf)
    reformatted = output_buf.getvalue()
    assert reformatted.endswith("\n")
    reformatted_output = problem.parse_output(input, io.StringIO(reformatted))
    reformatted_output.verify()
    assert output.score == reformatted_output.score


def test_bad_output(problem, bad_output_path, ascii_format):
    input_path, output_path = bad_output_path
    input = problem.parse_input(stringio_from_path(input_path))
    with pytest.raises(problemlib.FileFormatError):
        problem.parse_output(
            input, stringio_from_path(output_path, ascii_format=ascii_format)
        )


def test_rejected_output(problem, rejected_output_path, ascii_format):
    input_path, output_path = rejected_output_path
    input = problem.parse_input(stringio_from_path(input_path))
    output = problem.parse_output(
        input, stringio_from_path(output_path, ascii_format=ascii_format)
    )
    with pytest.raises(problemlib.VerificationError):
        output.verify()

    # Reformatting an output should still result in a rejected output.
    output_buf = io.StringIO()
    output.write(output_buf)
    reformatted = output_buf.getvalue()
    assert reformatted.endswith("\n")
    reformatted_output = problem.parse_output(input, io.StringIO(reformatted))
    with pytest.raises(problemlib.VerificationError):
        output.verify()
    assert output.score == reformatted_output.score


def test_generate_input(problem, rng):
    try:
        input = problem.generate_input(rng)
    except NotImplementedError:
        pytest.skip("Input.generate() is not required (yet!)")

    # The input should be writable.
    input_buf = io.StringIO()
    input.write(input_buf)

    # We should be able to parse the generated input.
    problem.parse_input(io.StringIO(input_buf.getvalue()))

    # Solve the generated input.
    try:
        output = input.trivial_solve()
    except NotImplementedError:
        pytest.skip("Input.trivial_solve() not implemented")

    # The solved output should be valid.
    output.verify()

    # The solved output should be writable.
    output_buf = io.StringIO()
    output.write(output_buf)

    # We should be able to parse the written output.
    problem.parse_output(input, io.StringIO(output_buf.getvalue()))


def load_inputs_from_dir(path):
    return list(path.glob("*.in"))


def load_outputs_from_dir(path):
    result = []
    for input_path in load_inputs_from_dir(path):
        assert input_path.is_symlink()
        output_path = input_path.parent / f"{input_path.stem}.out"
        result.append((input_path, output_path))
    return result


def run_problem_tests(problem_dir, pytest_extra_args=()):
    problem_dir = pathlib.Path(problem_dir).resolve()
    test_data_dir = problem_dir / "test_data"

    good_inputs = load_inputs_from_dir(test_data_dir / "inputs" / "good")
    bad_inputs = load_inputs_from_dir(test_data_dir / "inputs" / "bad")

    good_outputs = load_outputs_from_dir(test_data_dir / "outputs" / "good")
    bad_outputs = load_outputs_from_dir(test_data_dir / "outputs" / "bad")
    rejected_outputs = load_outputs_from_dir(test_data_dir / "outputs" / "rejected")

    class ProblemDataPlugin:
        def pytest_generate_tests(self, metafunc):
            def _add_param(name, value):
                if name in metafunc.fixturenames:
                    metafunc.parametrize(name, value)

            _add_param("good_input_path", good_inputs)
            _add_param("bad_input_path", bad_inputs)
            _add_param("good_output_path", good_outputs)
            _add_param("bad_output_path", bad_outputs)
            _add_param("rejected_output_path", rejected_outputs)
            _add_param("problem_dir", [problem_dir])
            _add_param("ascii_format", list(ASCIIFormat))

    return pytest.main([*pytest_extra_args, __file__], plugins=[ProblemDataPlugin()])


def main(argv=sys.argv):
    sys.exit(run_problem_tests(argv[1], pytest_extra_args=argv[2:]))


if __name__ == "__main__":
    main()



from pathlib import Path

import setuptools

HERE = Path(__file__).resolve().parent

BASE_DEPENDS = [
    "click>=8.0",
    "pytest>=4.6",
    "pytest-cov>=4.1.0",
    "requests>=2.0",
    "tabulate>=0.8",
    "toml>=0.10",
    "typing-extensions",
]

WEB_DEPENDS = [
    "TurboGears2 >= 2.3.12",
    "tgext.admin-jmr",
    "decorator",
    "google-api-python-client",
    "google-auth-oauthlib",
    "tw2.forms",
    "Beaker >= 1.8.0",
    "Kajiki >= 0.6.3",
    "filedepot >= 0.6.0",
    "zope.sqlalchemy >= 1.2",
    "sqlalchemy>=1.3,<1.4",
    "alembic>=0.8.8",
    "repoze.who",
    "WebHelpers2",
    "recordclass",
    "requests",
]

DEV_DEPENDS = [
    "tg.devtools",
]

setuptools.setup(
    name="algobowl",
    version=(HERE / "VERSION").read_text(encoding="ascii").strip(),
    description="Competition-based group project for Algorithms courses",
    long_description=(HERE / "README.md").read_text(encoding="utf-8"),
    long_description_content_type="text/markdown",
    author="Jack Rosenthal",
    author_email="jack@rosenth.al",
    url="https://github.com/jackrosenthal/algobowl",
    packages=setuptools.find_packages(),
    python_requires=">=3.8,<4",
    install_requires=BASE_DEPENDS,
    extras_require={
        "web": WEB_DEPENDS,
        "dev": WEB_DEPENDS + DEV_DEPENDS,
    },
    include_package_data=True,
    package_data={"algobowl": ["templates/*/*", "public/*/*"]},
    entry_points={
        "console_scripts": [
            "algobowl = algobowl.cli.__main__:main",
        ],
        "paste.app_factory": ["main = algobowl.config.middleware:make_app"],
        "gearbox.plugins": ["turbogears-devtools = tg.devtools"],
    },
)



import base64
import getpass
import os
import platform
import sys
import urllib.parse

import click
import requests

import algobowl.cli.formatter as fmt
import algobowl.lib.constants


def get_session(config):
    session = requests.Session()
    auth_token = config.get_server_config("access_token")
    if auth_token and len(auth_token) == 88:
        session.headers.update({"Authorization": f"Bearer {auth_token}"})
    session.headers.update(
        {
            "Accept": "application/json",
            "X-AlgoBOWL-CLI-Compatible": str(algobowl.lib.constants.cli_compatible),
        }
    )
    return session


def get_tokens(cli):
    r = cli.session.get(cli.config.get_url("/pref/cli.json"))
    check_response(r)
    return r.json()["auth_tokens"]


def check_response(response):
    if response.status_code == requests.codes.unauthorized:
        fmt.err("This command requires authentication.")
        fmt.err(f"Run {fmt.cmd('algobowl auth login')} to continue.")
        sys.exit(1)
    if not response.ok:
        fmt.err(
            {
                requests.codes.forbidden: "You are not authorized to perform this command.",
            }.get(response.status_code, "Your request failed to process.")
        )
        try:
            data = response.json()
        except requests.exceptions.JSONDecodeError:
            data = {}
        fmt.err(f"Reason: {data.get('message', 'Unknown.')}")
        fmt.err(f"Error {response.status_code} on URL {response.url}.")
        sys.exit(1)


def generate_client_name():
    return f"{platform.system()} {getpass.getuser()}@{platform.node()}"


def generate_client_id():
    return base64.urlsafe_b64encode(os.urandom(66)).decode("ascii")


@click.group(help="Authentication")
def auth():
    pass


@auth.command(help="Connect this CLI to your account")
@click.pass_obj
def login(cli):
    client_name = generate_client_name()
    client_id = generate_client_id()
    params = urllib.parse.urlencode(
        {
            "client_name": client_name,
            "client_id": client_id,
        }
    )
    url = cli.config.get_url(f"/pref/cli?{params}")
    cli.config.set_server_config("access_token", client_id)
    click.echo("Open the following URL in your browser:")
    click.echo(f"    {url}")
    click.echo(f"Then, run {fmt.cmd('algobowl auth whoami')} to verify.")


@auth.command(help="Show information about the current user")
@click.pass_obj
def whoami(cli):
    r = cli.session.get(cli.config.get_url("/pref/whoami.json"))
    check_response(r)
    cli.formatter.dump_table([r.json()])


@auth.command(help="Show registered clients")
@click.pass_obj
def list_clients(cli):
    cli.formatter.dump_table(get_tokens(cli))


@auth.command(help="Revoke a client token")
@click.option(
    "--token-id", type=int, help="Revoke a specific token (instead of this client's)"
)
@click.pass_obj
def logout(cli, token_id):
    if not token_id:
        tokens = get_tokens(cli)
        for token in tokens:
            if token["this_client"]:
                token_id = token["id"]
                break

    r = cli.session.get(
        cli.config.get_url("/pref/revoke_auth_token"),
        allow_redirects=False,
        params={"token_id": token_id},
    )
    check_response(r)



import datetime
import io
import json
import random

import tg
from depot.io.utils import FileIntent

import algobowl.lib.base as base
import algobowl.lib.problem as problemlib
import algobowl.model as model


def get_user(username):
    user = (
        model.DBSession.query(model.User)
        .filter(model.User.username == username)
        .one_or_none()
    )
    if user:
        return user

    mpapi = tg.request.environ["repoze.who.plugins"]["mpapi"]
    return mpapi.new_user_by_username(username)


def generate_default_input(
    problem: problemlib.Problem,
    group: model.Group,
    rng: random.Random,
) -> model.Input:
    reformatted_contents = io.StringIO()
    iput = problem.generate_input(rng)
    iput.write(reformatted_contents)

    input_file = FileIntent(
        io.BytesIO(reformatted_contents.getvalue().encode("utf-8")),
        f"input_group{group.id}.txt",
        "application/octet-stream",
    )
    result = model.Input(data=input_file, group=group, is_default=True)
    model.DBSession.add(result)
    return result


class SetupController(base.BaseController):
    allow_only = tg.predicates.has_permission("admin")

    @tg.expose("algobowl.templates.setup")
    def index(self):
        return dict(page="setup")

    @tg.expose(content_type="text/plain")
    def setup_competition(
        self,
        name=None,
        problem=None,
        input_upload_ends=None,
        output_upload_hours=48,
        verification_hours=24,
        resolution_hours=24,
        ov_hours=72,
        teams_json=None,
    ):
        def _setup():
            input_upload_ends_dt = datetime.datetime.fromisoformat(input_upload_ends)
            output_upload_delta = datetime.timedelta(hours=int(output_upload_hours))
            verification_delta = datetime.timedelta(hours=int(verification_hours))
            resolution_delta = datetime.timedelta(hours=int(resolution_hours))
            ov_delta = datetime.timedelta(hours=int(ov_hours))
            teams_json_data = json.load(teams_json.file)

            output_upload_ends = input_upload_ends_dt + output_upload_delta
            verification_ends = output_upload_ends + verification_delta
            resolution_ends = verification_ends + resolution_delta
            ov_ends = resolution_ends + ov_delta

            yield b"Setting up competition, please be patient\n"

            competition = model.Competition(
                name=name,
                problem=problem,
                allow_custom_team_names=True,
                input_upload_begins=input_upload_ends_dt - datetime.timedelta(weeks=2),
                input_upload_ends=input_upload_ends_dt,
                output_upload_begins=input_upload_ends_dt,
                output_upload_ends=output_upload_ends,
                verification_begins=output_upload_ends,
                verification_ends=verification_ends,
                resolution_begins=verification_ends,
                resolution_ends=resolution_ends,
                open_verification_begins=resolution_ends,
                open_verification_ends=ov_ends,
                evaluation_begins=resolution_ends,
                evaluation_ends=ov_ends,
            )

            yield b"Loading problem\n"
            # Validate problem name
            problem_obj = problemlib.load_problem(competition)
            problem_obj.get_module()
            model.DBSession.add(competition)

            rng = random.SystemRandom()
            for i, team in enumerate(teams_json_data, 1):
                yield f"Setting up team {i}/{len(teams_json_data)}\n".encode("utf-8")
                with model.DBSession.no_autoflush:
                    users = [get_user(username) for username in team if username]
                if not users:
                    continue
                group = model.Group(users=users, competition=competition)
                model.DBSession.add(group)
                group.input = generate_default_input(
                    problem=problem_obj,
                    group=group,
                    rng=rng,
                )

            yield b"Setting up admins team\n"
            with model.DBSession.no_autoflush:
                admins = (
                    model.DBSession.query(model.User)
                    .filter(model.User.admin == True)
                    .all()
                )
            group = model.Group(
                users=admins,
                competition=competition,
                name="Your Instructors",
            )
            model.DBSession.add(group)
            group.input = generate_default_input(
                problem=problem_obj,
                group=group,
                rng=rng,
            )

            yield b"Flushing database transactions\n"
            model.DBSession.flush()

            yield b"Completed successfully!\n"

        return _setup()



import dataclasses
import pathlib

import click
import requests

import algobowl.cli.admin as admin
import algobowl.cli.auth as auth
import algobowl.cli.config as cfg
import algobowl.cli.formatter as fmt
import algobowl.cli.group as group
import algobowl.cli.problem as problem


@dataclasses.dataclass
class CommandContext:
    config: cfg.CLIConfig = None
    session: requests.Session = None
    formatter: fmt.JsonFormatter = dataclasses.field(default_factory=fmt.TableFormatter)
    selected_group: int = -1


@click.group(help="The AlgoBOWL CLI")
@click.option("--config", type=pathlib.Path, help="Config file to use")
@click.option("--server", help="Server to use")
@click.option("--json", is_flag=True, help="Use JSON output instead of tables")
@click.pass_context
def main(ctx, server, config, json):
    ctx.ensure_object(CommandContext)

    if not config:
        config = cfg.get_default_config_path()

    ctx.obj.config = cfg.CLIConfig(path=config, server=server)
    ctx.obj.session = auth.get_session(ctx.obj.config)
    if json:
        ctx.obj.formatter = fmt.JsonFormatter()
    ctx.call_on_close(ctx.obj.config.write)


main.add_command(auth.auth)
main.add_command(cfg.config)
main.add_command(group.group)
main.add_command(admin.admin)
main.add_command(problem.problem)


if __name__ == "__main__":
    main()
